/******************************************************************************
*
* CHECK_CPU.C
*
* Program: Monitor cpu consumption plugin for Nagios
* License: GPL
* Copyright (c) 2002 Dave Viner <dviner@apache.org>
*
* Last Modified: $Date: 2002/11/13 11:50:54 $
*
* Description:
*
* This plugin will check for processes whose cpu consumption exceeds
* the warning or critical thresholds given on the command line.   With
* no command_name, everything that shows up on ps is evaluated.
* Otherwise, only jobs with the command_name given are examined.
* This program is particularly useful if you have to run a piece of
* commercial software that has a memory leak.  With it you can shut
* down and restart the processes whenever the program threatens to
* take over your system.
*
* Modifications:
*
*****************************************************************************/

#define PROGNAME "check_cpu"
#define REVISION "$Revision: 1.4 $"
#define COPYRIGHT "1999-2002"
#define AUTHOR "Dave Viner"
#define EMAIL "dviner@apache.org"
#define SUMMARY "Check the cpu consumption of a running program.\n"

#include "common.h"
#include "popen.h"
#include "utils.h"


// Required for systems without INFTIM defined, eg Solaris
#ifndef INFTIM
#define INFTIM (-1)
#endif

struct cmd_membuf {
	char *stdout_text;
	char *stderr_text;
	int stdout_len;
	int stderr_len;
};

int process_arguments (int argc, char **argv);
void print_help (char *cmd);
void print_usage (char *cmd);
int runcmd (char *, int , struct cmd_membuf *);

int examine_cpu_cons(int , char **, float , char *);
static char* getToken(char** , const char*);

int warn = -1;
int crit = -1;
char *proc = NULL;
int debug = FALSE;

int
main (int argc, char **argv)
{
	int result = STATE_OK;
	float cpu_cons;
	char proc_name[MAX_INPUT_BUFFER];
	char *message = "";
	char *cpy = NULL;
	char *textln = NULL;
	struct cmd_membuf cmd_out;

	if (!process_arguments (argc, argv)) {
		printf ("%s: failure parsing arguments\n", my_basename (argv[0]));
		print_help (my_basename (argv[0]));
		return STATE_UNKNOWN;
	}
	
	/* run the command */
	result = runcmd (CPU_COMMAND, timeout_interval, &cmd_out);
	 
	if (cmd_out.stdout_len == 0)
	{
		printf ("No output to stdout!\n");
		result = STATE_UNKNOWN;
	}
	else
	{
		/* 
		* the output from ps should be on stdout, so examine it line by line 
		* but let's keep the original around in case anyone wants it later
		*/
		cpy = strdup(cmd_out.stdout_text); 
		for ( ; (textln = getToken(&cpy, "\n")) != NULL; )
		{
			bzero (proc_name, MAX_INPUT_BUFFER);
			cpu_cons = 0;
			if (sscanf (textln, CPU_FORMAT, &cpu_cons, proc_name) == 2) 
			{
				/* examine the size information */
				result = examine_cpu_cons (result, &message, cpu_cons, proc_name);
			}
		}
		/* free up the copy we made */
		free (cpy);

	}
	
	

	if (result == STATE_OK)
		printf ("ok (all CPU<%d): %s\n", warn, message);
	else if (result == STATE_UNKNOWN)
		printf ("Unable to read output\n");
	else if (result == STATE_WARNING)
		printf ("WARNING (CPU>%d):%s\n", warn, message);
	else
		printf ("CRITICAL (CPU>%d):%s\n", crit, message);
	
	
	return result;
}

/*
* look at the consumption passed in, and see if we need to 
* add to the message that we'll report in the end
*/
int
examine_cpu_cons(int result, char **message, float cpu_cons, char *proc_name)
{
	if (proc == NULL) {
		if (cpu_cons > warn) {
			asprintf (message, "%s %s(%.1f)", *message, proc_name, cpu_cons);
			result = max_state (result, STATE_WARNING);
		}
		if (cpu_cons > crit) {
			result = STATE_CRITICAL;
		}
	}
	else if (strstr (proc_name, proc)) {
		asprintf (message, "%s %.1f", *message, cpu_cons);
		if (cpu_cons > warn) {
			result = max_state (result, STATE_WARNING);
		}
		if (cpu_cons > crit) {
			result = STATE_CRITICAL;
		}
	}
	return result;
}



#define BUFFER_SIZE 8*1024
/*
* This little subrountine will take a command to spopen, and
* the timeout requested, and a pointer to the cmd_membuf.
* It will collect all the stdout output into membuf->stdout_text
* and the number of bytes written there into membuf->stdout_len.
* It will also collection all the stderr output into membuf->stderr_text
* and the number of bytes written there into membuf->stderr_len.
*
* N.B. - the stdout_text and stderr_text may NOT be null terminated.
* So, don't use printf to print them, use write().
*
* N.B. - this function will allocate memory for membuf as required.
* If you are concerned about memory usage, be sure to free the 
* membuf->stdout_text and membuf->stderr_text.
*/
int 
runcmd (char *cmd, int timeout_interval, struct cmd_membuf *membuf)
{
	int result = STATE_UNKNOWN;
	/* integers that are helpful */
	int n, read_size, child_stderr_num, child_stdout_num;
	struct pollfd fdarray[2];
	short readable = POLLIN;
	int stdout_resid = 0, stderr_resid = 0;
	int stdout_bufsz = 0, stderr_bufsz = 0;
	
	
	/* set the output memory size */
	membuf->stdout_len = 0;
	membuf->stderr_len = 0;
	membuf->stdout_text = NULL;
	membuf->stderr_text = NULL;

	/* Set signal handling and alarm timeout */	
	if (timeout_interval != 0)
	{
		if (signal (SIGALRM, popen_timeout_alarm_handler) == SIG_ERR) {
			printf ("Cannot catch SIGALRM");
			return STATE_UNKNOWN;
		}
		alarm (timeout_interval);
	}
	
	
	/* open the child process */
	if (debug)
		printf("Executing: %s\n", cmd);
	child_process = spopen (cmd);
	if (child_process == NULL) {
		printf ("Could not open pipe: %s\n", cmd);
		exit (STATE_UNKNOWN);
	}
	
	
	/* fetch the child file descriptors that we're interested in */
	child_stdout_num = fileno (child_process);
	child_stderr_num = child_stderr_array[child_stdout_num];
	
	if (debug)
		printf ("Child stdout is: %d, stderr is %d\n", 
		child_stdout_num, child_stderr_num);
	
	/* set up the poll array */
	fdarray[0].fd = child_stdout_num;
	fdarray[0].events = readable;
	fdarray[1].fd = child_stderr_num;
	fdarray[1].events = readable;
	
	while ( 1 )
	{
		
		n = poll (fdarray, 2, INFTIM); 
		if (n == -1)
		{
			printf ("ERROR from poll\n");
			result = STATE_UNKNOWN;
			break;
		}
		
		
		/* 
		* try reading from stdout 
		*/
		if ((fdarray[0].revents & readable))
		{
			/* see if we need to resize the stdout buffer */
			if (stdout_resid == 0) {
				stdout_resid = BUFFER_SIZE;
				stdout_bufsz += stdout_resid;
				membuf->stdout_text = realloc (membuf->stdout_text, 
					stdout_bufsz);
			}
			
			read_size = read (child_stdout_num, 
							  membuf->stdout_text + membuf->stdout_len,
							  stdout_bufsz);
			membuf->stdout_len += read_size;
			stdout_resid -= read_size;
			
			/* 
			* if 0 length is returned, then we're finished reading from 
			* child process
			*/
			if (read_size == 0)
				break;
		}
		
		/* 
		* try reading from stderr 
		*/
		if ((fdarray[1].revents & readable))
		{
			/* see if we need to resize the stderr buffer */
			if (stderr_resid == 0) {
				stderr_resid = BUFFER_SIZE;
				stderr_bufsz += stderr_resid;
				membuf->stderr_text = realloc(membuf->stderr_text, stderr_bufsz);
			}
			
			read_size = read (child_stderr_num, membuf->stderr_text + membuf->stderr_len, stderr_bufsz);
			membuf->stderr_len += read_size;
			stderr_resid -= read_size;
		}
		
		/* Solaris needs this.  It will send POLLHUP when all is through
		* FreeBSD simply has a read size of 0 from the stdout.
		*/
		if ((fdarray[0].revents & (POLLHUP|POLLIN)) == POLLHUP || 
			(fdarray[1].revents & (POLLHUP|POLLIN)) == POLLHUP)
			break;
		
	} /* while ( 1 ) */
	
	/* close child */
	result = spclose (child_process); 
	
	if (debug)
	{
		write(1, "STDOUT: ", 9);
		if(membuf->stdout_len > 0)
		{
			write (1, membuf->stdout_text, membuf->stdout_len);
			printf("\n");
		}
		else
			printf("(no output)\n");
		
		write(1, "STDERR: ", 9);
		if(membuf->stderr_len > 0)
		{
			write (1, membuf->stderr_text, membuf->stderr_len);
			printf("\n");
		}
		else
			printf("(no output)\n");
	}
	
	return result;
}


int
process_arguments (int argc, char **argv)
{
	int c;
	
#ifdef HAVE_GETOPT_H
	int option_index = 0;
	static struct option long_options[] = {
		{"help", no_argument, 0, 'h'},
		{"version", no_argument, 0, 'V'},
		{"critical", required_argument, 0, 'c'},
		{"warning", required_argument, 0, 'w'},
		{"command", required_argument, 0, 'C'},
		{"debug", no_argument, 0, 'd'},
		{"timeout", required_argument, 0, 't'},
		{0, 0, 0, 0}
	};
#endif
	
	if (argc < 2)
		return ERROR;
	
	while (1) {
#ifdef HAVE_GETOPT_H
		c = getopt_long (argc, argv, "+hdVc:w:C:t:", long_options, &option_index);
#else
		c = getopt (argc, argv, "+hdVc:w:C:t:");
#endif
		if (c == EOF)
			break;
		
		switch (c) {
		case '?':									/* help */
			printf ("%s: Unknown argument: %s\n\n", my_basename (argv[0]), optarg);
			print_usage (my_basename (argv[0]));
			exit (STATE_UNKNOWN);
		case 'h':									/* help */
			print_help (my_basename (argv[0]));
			exit (STATE_OK);
		case 'V':									/* version */
			print_revision (my_basename (argv[0]), "$Revision: 1.4 $");
			exit (STATE_OK);
		case 'c':									/* critical threshold */
			if (!is_intnonneg (optarg)) {
				printf ("%s: critical threshold must be an integer: %s\n",
					my_basename (argv[0]), optarg);
				print_usage (my_basename (argv[0]));
				exit (STATE_UNKNOWN);
			}
			crit = atoi (optarg);
			break;
		case 'w':									/* warning threshold */
			if (!is_intnonneg (optarg)) {
				printf ("%s: warning threshold must be an integer: %s\n",
					my_basename (argv[0]), optarg);
				print_usage (my_basename (argv[0]));
				exit (STATE_UNKNOWN);
			}
			warn = atoi (optarg);
			break;
		case 'C':									/* command name */
			proc = optarg;
			break;
		case 't':
			timeout_interval = atoi (optarg);
			break;
		case 'd':
			debug = TRUE;
			break;
		}
	}
	
	for ( c = optind; c < argc; c++ )
	{
		if (warn == -1) {
			if (!is_intnonneg (argv[c])) {
				printf ("%s: critical threshold must be an integer: %s\n",
					PROGNAME, argv[c]);
				print_usage (PROGNAME);
				exit (STATE_UNKNOWN);
			}
			warn = atoi (argv[c]);
		}
		else
			if (crit == -1) {
				if (!is_intnonneg (argv[c])) {
					printf ("%s: critical threshold must be an integer: %s\n",
						PROGNAME, argv[c]);
					print_usage (PROGNAME);
					exit (STATE_UNKNOWN);
				}
				crit = atoi (argv[c]);
			}
			else
				if (proc == NULL)
					proc = argv[c];
	}
	
	return c;
}

void
print_usage (char *cmd)
{
	printf ("Usage: %s -w <wsize> -c <csize> [-C command]\n"
		"       %s --help\n" "       %s --version\n", cmd, cmd, cmd);
}

void
print_help (char *cmd)
{
	print_revision (PROGNAME, REVISION);
	printf
		("Copyright (c) 2002 Dave Viner <dviner@apache.org>\n\n"
		"This plugin checks the image size of a running program and returns an\n"
		"error if the number is above either of the thresholds given.\n\n");
	print_usage (cmd);
	printf
		("\nOptions:\n"
		" -h, --help\n"
		"    Print detailed help\n"
		" -V, --version\n"
		"    Print version numbers and license information\n"
		" -w, --warning=INTEGER\n"
		"    Program image size necessary to cause a WARNING state\n"
		" -c, --critical=INTEGER\n"
		"    Program image size necessary to cause a CRITICAL state\n"
		" -C, --command=STRING\n" "    Program to search for [optional]\n");
}

/***************************************************************************
* This is a replacement for strsep which is not portable (missing on Solaris).
* From Wine code by Francois Gouget <fgouget@codeweavers.com>
*/
static char* getToken(char** str, const char* delims)
{
	char* token;
	
	if (*str==NULL) {
		/* No more tokens */
		return NULL;
	}
	
	token=*str;
	while (**str!='\0') {
		if (strchr(delims,**str)!=NULL) {
			**str='\0';
			(*str)++;
			return token;
		}
		(*str)++;
	}
	/* There is no other token */
	*str=NULL;
	return token;
}

